%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                    %
%     An Automatically Verified Prototype of the     %
%                                                    %
%         Tokeneer ID Station Specification          %
%                                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

optional(X,Opt) :- subset(Opt,X) & subset(Opt,{_}).
optionalBasic(Opt) :- subset(Opt,{_}).
optionalSchema(Opt,IsOpt) :- subset(Opt,{IsOpt}).

% the(Type,Set,Elem) :- subset(Set,Type) & Set = {Elem}.  
% the({Elem},Elem).  
the(Set,Elem) :- Set = {Elem}.  

zeroTime(T) :- T = 0.

class(CLASS) :- 
  CLASS = {unmarked, unclassified, restricted, confidential, secret, topsecret}.

privilege(PRIVILEGE) :- 
  PRIVILEGE = {userOnly, guard, securityOfficer, auditManager}.

adminprivilege(ADMINPRIVILEGE) :- 
  ADMINPRIVILEGE = {guard, securityOfficer, auditManager}.

minClearance(_C1,_C2,_Min_C1_C2).

caIdCertPred(IDCert) :- 
  IDCert = [[_,_,IsValidatedBy],[_,SubjectPubK]] &

  IsValidatedBy = {SubjectPubK}.

validToken(Token) :-
  Token = [TokenID,IDCert,PrivCert,IandACert,_] & 
    PrivCert = [_,_,_,PBaseCertId,PTokenID,_,_] &
    IDCert = [[Id,_,_],[_,_]] &
    IandACert = [_,_,_,IBaseCertId,ITokenID,_] &

  PBaseCertId = Id &
  IBaseCertId = Id &
  PTokenID = TokenID &
  ITokenID = TokenID.

%  AuthCert neq {} unnecessary, the(AuthCert,TheAuthCert) won't unify
tokenWithValidAuth(Token) :-
  Token = [TokenID,IDCert,_,_,AuthCert] &
    IDCert = [[Id,_,_],[_,_]] &

  the(AuthCert,TheAuthCert) &
    TheAuthCert = [_,_,_,BaseCertId,TTokenID,_,_] &
  TTokenID = TokenID &
  BaseCertId = Id.

currentToken(Token,Now) :-
  Token = [_,IDCert,PrivCert,IandACert,_] &
    IDCert = [[_,ValidityPeriod,_],[_,_]] &
    PrivCert = [_,PValidityPeriod,_,_,_,_,_] &
    IandACert = [_,IValidityPeriod,_,_,_,_] &

  validToken(Token) &

  0 =< Now &   % implicit precondition

  inters(ValidityPeriod,PValidityPeriod,M1) &
  inters(M1,IValidityPeriod,{Now/_}).

% issuerCerts \cap \{CAIdCert\} \neq \emptyset
% there's at least one element in issuerCerts that's
% a CAIdCert, i.e. it's an IDCert verifying the invariant
validEnrol(Enrol) :-
  Enrol = [IdStationCert,IssuerCerts] &

  IdStationCert in IssuerCerts &

  IssuerCerts = {IDCert/_} & caIdCertPred(IDCert) &
  delay(
    foreach(
      [[[Issuer],_,IsValidatedBy],[_,_]] in IssuerCerts,
      [SPK],
      IsValidatedBy neq {} &
      ris([[_,_,IsValidatedBy1],[Subject1,SubjectPubK1]] 
            in IssuerCerts,
          IsValidatedBy1 = {SubjectPubK1} &
          SPK = SubjectPubK1 &
          Issuer = Subject1
         ) neq {},
      the(IsValidatedBy,SPK)
    ),
    false
  ).

maxSupportedLogSize(MaxSupportedLogSize) :- 0 =< MaxSupportedLogSize.

sizeElement(SizeElement) :-
  pfun(SizeElement) &
  foreach([_X,Y] in SizeElement, 0 =< Y).

sizeLog(SizeLog) :-
  pfun(SizeLog) &
  foreach([_X,Y] in SizeLog, 0 =< Y).

sizeLogProp1(SizeLog) :-
  sizeLog(SizeLog) &
  [{},0] in SizeLog.

% We can't define sizeLog as in the Z specification
% all we can do is to state which is the sizeLog of a singleton...
sizeLogProp2(SizeLog) :-
  sizeLog(SizeLog) &
  [{A},M] in SizeLog & sizeElement(SE) & [A,M] in SE.

% ... and that sizeLog is monotone
sizeLogProp3(SizeLog) :-
  sizeLog(SizeLog) &
  foreach([A1,N1] in SizeLog,
    foreach([A2,N2] in SizeLog,
      nsubset(A1,A2) or N1 =< N2)).  % A1 \subseteq A2 \implies N1 \leq N2

% oldestLogTime is a decreasing function
% we can't express it as in the Z specification
% we instead express it with a double RUQ
oldestLogTime(OldestLogTime) :-
  pfun(OldestLogTime) &
  foreach([_A,N] in OldestLogTime, 0 =< N) &
  foreach([A,M] in OldestLogTime,
    foreach([B,N] in OldestLogTime,
      nsubset(A,B) or N =< M)).

% newestLogTime is an increasing function
% we can't express it as in the Z specification
% we instead express it with a double RUQ
newestLogTime(NewestLogTime) :-
  pfun(NewestLogTime) &
  foreach([_A,N] in NewestLogTime, 0 =< N) &
  foreach([A,M] in NewestLogTime,
    foreach([B,N] in NewestLogTime,
      nsubset(A,B) or M =< N)).

realWorldChanges(RealWorld,RealWorld_) :-
  RealWorld = [_TISControlledRealWorld,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [Now,_,_,_,_,_,_] &
  RealWorld_ = [_TISControlledRealWorld_,TISMonitoredRealWorld_] &
    TISMonitoredRealWorld_ = [Now_,_,_,_,_,_,_] &

  Now_ >= Now.

pollTime(RealWorld,_DoorLatchAlarm,DoorLatchAlarm_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [Now,_,_,_,_,_,_] &
  DoorLatchAlarm_ = [CurrentTime_,_,_,_,_,_] &

  CurrentTime_ = Now.

pollDoor(RealWorld,DoorLatchAlarm,DoorLatchAlarm_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,Door,_,_,_,_,_] &
  DoorLatchAlarm = [_,_,_,_,LatchTimeout,AlarmTimeout] &
  DoorLatchAlarm_ = [_,CurrentDoor_,_,_,LatchTimeout_,AlarmTimeout_] &

  CurrentDoor_ = Door &
  LatchTimeout_ = LatchTimeout &
  AlarmTimeout_ = AlarmTimeout.

pollUserToken(RealWorld,UserToken,UserToken_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,_,_,UserTokenRW,_,_,_] &
  UserToken = [CurrentUserToken,_] &
  UserToken_ = [CurrentUserToken_,UserTokenPresence_] &

  (UserTokenRW = noT or UserTokenPresence_ = present) &
  (UserTokenPresence_ neq present or UserTokenRW neq noT) &
  (UserTokenRW = noT or CurrentUserToken_ = UserTokenRW) &
  (UserTokenRW neq noT or CurrentUserToken_ = CurrentUserToken).

pollAdminToken(RealWorld,AdminToken,AdminToken_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,_,_,_,AdminTokenRW,_,_] &
  AdminToken = [CurrentAdminToken,_] &
  AdminToken_ = [CurrentAdminToken_,AdminTokenPresence_] &

  (AdminTokenRW = noT or AdminTokenPresence_ = present) &
  (AdminTokenPresence_ neq present or AdminToken neq noT) &
  (AdminTokenRW = noT or CurrentAdminToken_ = AdminTokenRW) &
  (AdminTokenRW neq noT or CurrentAdminToken_ = CurrentAdminToken) &

  % implicit condition; can do better
  tokenWithValidAuth(T) & goodT(T) = CurrentAdminToken_.

pollFinger(RealWorld,Finger,Finger_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,_,Finger,_,_,_,_] &
  Finger = [CurrentFinger,_] &
  Finger_ = [CurrentFinger_,FingerPresence_] &

  (Finger = noFP or FingerPresence_ = present) &
  (FingerPresence_ neq present or Finger neq noFP) &
  (Finger = noFP or CurrentFinger_ = Finger) &
  (Finger neq noFP or CurrentFinger_ = CurrentFinger).

pollFloppy(RealWorld,Floppy,Floppy_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,_,_,_,_,Floppy,_] &
  Floppy = [CurrentFloppy,WrittenFloppy,_] &
  Floppy_ = [CurrentFloppy_,WrittenFloppy_,FloppyPresence_] &

  (Floppy = noFloppy or FloppyPresence_ = present) &
  (FloppyPresence_ neq present or Floppy neq noFloppy) &
  (Finger = noFloppy or CurrentFloppy_ = Floppy) &
  (Finger neq noFloppy or CurrentFloppy_ = CurrentFloppy) &
  WrittenFloppy_ = WrittenFloppy.

% possible mistake in:
%  (Keyboard = noKB or KeyedDataPresence = present) &
%  (KeyedDataPresence neq present or Keyboard neq noKB) &
% because there's no KeyedDataPresence_
pollKeyboard(RealWorld,Keyboard,Keyboard_) :-
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [_,_,_,_,_,_,Keyboard] &
  Keyboard = [CurrentKeyedData,KeyedDataPresence] &
  Keyboard_ = [CurrentKeyedData_,_] &

  (Keyboard = noKB or KeyedDataPresence = present) &
  (KeyedDataPresence neq present or Keyboard neq noKB) &
  (Keyboard = noKB or CurrentKeyedData_ = Keyboard) &
  (Keyboard neq noKB or CurrentKeyedData_ = CurrentKeyedData).

tisPoll(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,DoorLatchAlarm,Floppy,
     Keyboard,Config,Stats,KeyStore,Admin,_,Internal,
     CurrentDisplay,CurrentScreen] &
    DoorLatchAlarm = [_,_,_,_,LatchTimeout,AlarmTimeout] &
    Internal = [Status,_,_] &
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [Now,Door,_,_,_,_,_] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,Stats_,KeyStore_,Admin_,_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    UserToken_ = [CurrentUserToken_,_] &
    DoorLatchAlarm_ = [_,_,CurrentLatch_,DoorAlarm_,_,_] &

  RealWorld_ = RealWorld &

  pollTime(RealWorld,DoorLatchAlarm,DoorLatchAlarm_) &
  pollDoor(RealWorld,DoorLatchAlarm,DoorLatchAlarm_) &
    % implicit condition
    (Now >= LatchTimeout &  % CurrentTime_ = Now in pollTime
     CurrentLatch_ = locked
     or
     Now < LatchTimeout &
     CurrentLatch_ = unlocked
    ) &
    (Door = open &  % CurrentDoor_ = Door in pollDoor
     Now >= LatchTimeout &  % CurrentLatch_ = locked
     Now >= AlarmTimeout &  % CurrentTime_ = Now in pollTime
     DoorAlarm_ = alarming
     or
     (Door = closed
      or
      Now < LatchTimeout
      or
      Now < AlarmTimeout
     ) & DoorAlarm_ = silent
    ) &
  pollUserToken(RealWorld,UserToken,UserToken_) &
    % implicit condition; can do better by breaking the design
    (Status nin {gotFinger, waitingFinger, waitingUpdateToken, waitingEntry}
     or 
     validToken(T) & goodT(T) = CurrentUserToken_
     or
     tokenWithValidAuth(T) & goodT(T) = CurrentUserToken_
    ) &
  pollAdminToken(RealWorld,AdminToken,AdminToken_) &
  pollFinger(RealWorld,Finger,Finger_) &
  pollFloppy(RealWorld,Floppy,Floppy_) &
  pollKeyboard(RealWorld,Keyboard,Keyboard_) &
  logChange(IDStation,IDStation_) &

  Config_ = Config &
  KeyStore_ = KeyStore &
  Admin_ = Admin &
  Stats_ = Stats &
  Internal_ = Internal &

  CurrentScreen_ = CurrentScreen &

  (CurrentDisplay = doorUnlocked &
   CurrentLatch_ = locked &
   (Status neq waitingRemoveTokenFail &
    CurrentDisplay_ = welcome
    or
    Status = waitingRemoveTokenFail &
    CurrentDisplay_ = removeToken
   ) 
   or
   CurrentDisplay_ = CurrentDisplay &
   (CurrentDisplay neq doorUnlocked 
    or
    CurrentLatch_ neq locked
   )
  ).

updateLatch(DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) :-
  DoorLatchAlarm = [_,_,CurrentLatch,_,_,_] &

  DoorLatchAlarm_ = DoorLatchAlarm &

  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld_ = [TISControlledRealWorld_,_] &
      TISControlledRealWorld_ = [Latch_,_,_,_] &

  Latch_ = CurrentLatch.

updateAlarm(AuditLog,DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) :-
  DoorLatchAlarm = [_,_,_,DoorAlarm,_,_] &
  AuditLog = [_,AuditAlarm] &

  DoorLatchAlarm_ = DoorLatchAlarm &

  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld_ = [TISControlledRealWorld_,_] &
      TISControlledRealWorld_ = [_,Alarm_,_,_] &

   % Alarm_ neq alarming <==> Alarm_ = silent
  (Alarm_ = silent or DoorAlarm = alarming or AuditAlarm = alarming) &
  (DoorAlarm neq alarming & AuditAlarm neq alarming or Alarm_ = alarming).

updateDisplay(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,_,_,_,_,_,_,CurrentDisplay,_] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,_,_,_,CurrentDisplay_,_] &

  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld_ = [TISControlledRealWorld_,_] &
      TISControlledRealWorld_ = [_,_,Display_,_] &

  Display_ = CurrentDisplay &
  CurrentDisplay_ = CurrentDisplay.

updateScreen(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,_,_,CurrentScreen] &
    CurrentScreen = [ScreenStats,ScreenMsg,ScreenConfig] &
    Admin = [RolePresent,_,_] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,Admin_,_,_,_,_] &

  Admin_ = Admin &
  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld_ = [TISControlledRealWorld_,_] &
      TISControlledRealWorld_ = [_,_,_,TISScreen_] &
      TISScreen_ = [TISScreenStats_,TISScreenMsg_,TISScreenConfig_] &

  TISScreenMsg_ = ScreenMsg &
  (the(RolePresent,R) & R neq securityOfficer 
   or 
   TISScreenConfig_ = ScreenConfig
  ) &
  (the(RolePresent,securityOfficer) 
   or 
   TISScreenConfig_ = clear
  ) &
  (RolePresent = {} 
   or 
   TISScreenStats_ = ScreenStats
  ) &
  (RolePresent neq {} 
   or 
   TISScreenStats_ =  clear
  ).

tisEarlyUpdate(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,DoorLatchAlarm,Floppy,
     Keyboard,Config,Stats,KeyStore,Admin,AuditLog,Internal,
     CurrentDisplay,CurrentScreen] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,Stats_,KeyStore_,Admin_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &
  RealWorld = [TISControlledRealWorld,_] &
    TISControlledRealWorld = [_,_,Display,Screen] &
  RealWorld_ = [TISControlledRealWorld_,_] &
    TISControlledRealWorld_ = [_,_,Display_,Screen_] &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  updateLatch(DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) &
  updateAlarm(AuditLog,DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) &

  realWorldChanges(RealWorld,RealWorld_) & Screen_=Screen & Display_=Display &

  CurrentDisplay_ = CurrentDisplay &

  UserToken_ = UserToken &
  AdminToken_ = AdminToken &
  Finger_ = Finger &
  Floppy_ = Floppy &
  Keyboard_ = Keyboard &
  Config_ = Config &
  Stats_ = Stats &
  KeyStore_ = KeyStore &
  Admin_ = Admin &
  Internal_ = Internal &

  (delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false)
   or 
   AuditLog_ = AuditLog
  ) &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisUpdate(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,DoorLatchAlarm,Floppy,
     Keyboard,Config,Stats,KeyStore,Admin,AuditLog,Internal,_,CurrentScreen] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,Stats_,KeyStore_,Admin_,AuditLog_,Internal_,_,
     CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  updateLatch(DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) &
  updateAlarm(AuditLog,DoorLatchAlarm,RealWorld,DoorLatchAlarm_,RealWorld_) &
  updateDisplay(IDStation,RealWorld,RealWorld_,IDStation_) &
  updateScreen(IDStation,RealWorld,RealWorld_,IDStation_) &
  UserToken_ = UserToken &
  AdminToken_ = AdminToken &
  Finger_ = Finger &
  Floppy_ = Floppy &
  Keyboard_ = Keyboard &
  Config_ = Config &
  Stats_ = Stats &
  KeyStore_ = KeyStore &
  Admin_ = Admin &
  Internal_ = Internal &
  (delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false)
   or 
   AuditLog_ = AuditLog
  ) &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

updateUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [UserToken,_,_,_,_,_,_,_,_,_,_,_,_,_] &
    UserToken = [CurrentUserToken,_] &

  IDStation_ = IDStation &
  realWorldChanges(RealWorld,RealWorld_) & 
    RealWorld = [TISControlledRealWorld,_] &
    RealWorld_ = [TISControlledRealWorld_,TISMonitoredRealWorld_] &
      TISMonitoredRealWorld_ = [_,_,_,TISUserToken_,_,_,_] &

  TISControlledRealWorld_ = TISControlledRealWorld &

  TISUserToken_ = CurrentUserToken.

updateFloppy(IDStation,RealWorld,IDStation_,RealWorld_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,DoorLatchAlarm,Floppy,
     Keyboard,Config,Stats,KeyStore,Admin,AuditLog,Internal,
     CurrentDisplay,CurrentScreen] &
    Floppy = [_,WrittenFloppy,FloppyPresence] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,Stats_,KeyStore_,Admin_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Floppy_ = [CurrentFloppy_,_,FloppyPresence_] &

  realWorldChanges(RealWorld,RealWorld_) & 
    RealWorld = [TISControlledRealWorld,_] &
    RealWorld_ = [TISControlledRealWorld,TISMonitoredRealWorld_] &
      TISMonitoredRealWorld_ = [_,_,_,_,_,TISFloppy_,_] &

  UserToken_ = UserToken &
  AdminToken_ = AdminToken &
  Finger_ = Finger &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Keyboard_ = Keyboard &
  Config_ = Config &
  Stats_ = Stats &
  KeyStore_ = KeyStore &
  Admin_ = Admin &
  AuditLog_ = AuditLog &
  Internal_ = Internal &

  TISFloppy_ = WrittenFloppy &

  CurrentFloppy_ = badFloppy &
  FloppyPresence_ = FloppyPresence &
  CurrentDisplay_ = CurrentDisplay &
  CurrentScreen_ = CurrentScreen.

% Instead of introducing an existensial quantifier,
% we put newElements as an argument (for instance, 
% as an input variable). When another operation that
% calls addElementsToLog knows what the elements to 
% be added to the audit log are (i.e. it has a value
% for newElements), then it simply pass that set
% as the argument; when the callee doesn't know the
% set, the argument acts as an existensial quantifier
addElementsToLog(Config,NewElements,AuditLog,AuditLog_) :-
  Config = [_,_,_,_,_,_,MinPreservedLogSize,AlarmThresholdSize] &
  AuditLog = [AauditLog,AuditAlarm] &
  AuditLog_ = [AauditLog_,AuditAlarm_] &

  NewElements neq {} &
  oldestLogTime(OT) & newestLogTime(NT) & sizeLog(SL) & 
  [NewElements,M1] in OT & [AauditLog,M2] in NT & M1 >= M2 &
  [AauditLog_,M3] in SL & 
  (un(AauditLog,NewElements,AauditLog_) &
   (M3 < AlarmThresholdSize &
    AuditAlarm_ = AuditAlarm
    or
    M3 >= AlarmThresholdSize &
    AuditAlarm_ = alarming
   )
   or
   [AauditLog,M4] in SL &
     [NewElements,M5] in SL & M4 + M5 > MinPreservedLogSize &
   un(OldElements,AauditLog_,M6) & un(AauditLog,NewElements,M6) &
   [AauditLog_,M7] in OT & [OldElements,M8] in NT & M7 >= M8 &
   M3 >= MinPreservedLogSize &
   AuditAlarm_ = alarming
  ).

% This operation calls addElementsToLog
% having existensially quantified a variable with the name
% newElements. With our encoding of addElementsToLog we
% simply pass in the quatified variable as the second
% argument to addElementsToLog
archiveLog(Config,AuditLog,Archive,AuditLog_) :-
  AuditLog = [AauditLog,_] &
  AuditLog_ = [AauditLog_,_] &

  subset(Archive,M1) & un(AauditLog,NewElements,M1) &
  subset(AauditLog_,M2) & un(Archive,NotArchived,M2) &
  M3 =< M4 &
    [Archive,M3] in NT & [NotArchived,M4] in OT & 
    newestLogTime(NT) & oldestLogTime(OT) & 
  delay(addElementsToLog(Config,NewElements,AuditLog,AuditLog_),false).

clearLog(Config,AuditLog,Archive,AuditLog_) :-
  Config = [_,_,_,_,_,_,_,AlarmThresholdSize] &
  AuditLog = [AauditLog,_] &
  AuditLog_ = [AauditLog_,AuditAlarm_] &

  oldestLogTime(OT) & newestLogTime(NT) & sizeLog(SL) & 
  un(Archive,SinceArchive,M1) & un(LostSinceArchive,AauditLog,M1) &
  [SinceArchive,M2] in OT & [Archive,M3] in NT & M2 >= M3 &
  [LostSinceArchive,M4] in NT & [AauditLog,M5] in OT & M4 =< M5 &
  AauditLog_ = SinceArchive &
  [AauditLog_,M6] in SL &
  (M6 < AlarmThresholdSize &
   AuditAlarm_ = silent
   or
   M6 >= AlarmThresholdSize &
   AuditAlarm_ = alarming
  ).

auditDoor(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_) :-
  DoorLatchAlarm = [_,CurrentDoor,_,_,_,_] &
  DoorLatchAlarm_ = [_,CurrentDoor_,_,_,_,_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  CurrentDoor_ neq CurrentDoor.

auditLatch(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_) :-
  DoorLatchAlarm = [_,_,CurrentLatch,_,_,_] &
  DoorLatchAlarm_ = [_,_,CurrentLatch_,_,_,_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  CurrentLatch_ neq CurrentLatch.

auditAlarm(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_) :-
  DoorLatchAlarm = [_,_,_,DoorAlarm,_,_] &
  DoorLatchAlarm_ = [_,_,_,DoorAlarm_,_,_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  DoorAlarm_ neq DoorAlarm.

auditLogAlarm(Config,AuditLog,AuditLog_) :-
  AuditLog = [_,AuditAlarm] &
  AuditLog_ = [_,AuditAlarm_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  AuditAlarm_ neq AuditAlarm.

auditDisplay(IDStation,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,AuditLog,_,CurrentDisplay,_] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,_,AuditLog_,_,CurrentDisplay_,_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  CurrentDisplay_ neq CurrentDisplay.

auditScreen(IDStation,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,AuditLog,_,_,CurrentScreen] &
    CurrentScreen = [_,ScreenMsg,_] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [_,ScreenMsg_,_] &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenMsg_ neq ScreenMsg.

noChange(IDStation,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,_,_,Config,_,_,_,AuditLog,_,
     CurrentDisplay,CurrentScreen] &
    DoorLatchAlarm = [_,CurrentDoor,CurrentLatch,DoorAlarm,_,_] &
    CurrentScreen = [_,ScreenMsg,_] &
    AuditLog = [_,AuditAlarm] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,_,_,_,_,_,_,AuditLog_,_,
     CurrentDisplay_,CurrentScreen_] &
    DoorLatchAlarm_ = [_,CurrentDoor_,CurrentLatch_,DoorAlarm_,_,_] &
    CurrentScreen_ = [_,ScreenMsg_,_] &
    AuditLog_ = [_,AuditAlarm_] &

  (CurrentDoor_ = CurrentDoor &
   CurrentLatch_ = CurrentLatch &
   DoorAlarm_ = DoorAlarm &
   AuditAlarm_ = AuditAlarm &
   CurrentDisplay_ = CurrentDisplay &
   ScreenMsg_ = ScreenMsg &
 
   delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) 
   or
   AuditLog_ = AuditLog
 ).

logChange(IDStation,IDStation_) :-
  IDStation = [_,_,_,DoorLatchAlarm,_,_,Config,_,_,_,AuditLog,_,_,_] &
  IDStation_ = [_,_,_,DoorLatchAlarm_,_,_,_,_,_,_,AuditLog_,_,_,_] &

  (auditAlarm(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_)
   or
   auditLatch(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_)
   or
   auditDoor(DoorLatchAlarm,Config,AuditLog,AuditLog_,DoorLatchAlarm_)
   or
   auditLogAlarm(Config,AuditLog,AuditLog_)
   or
   auditScreen(IDStation,IDStation_)
   or
   auditDisplay(IDStation,IDStation_)
   or
   noChange(IDStation,IDStation_)
  ).

addSuccessfulEntryToStats(Stats,Stats_) :-
  Stats = [SuccessEntry,FailEntry,SuccessBio,FailBio] &
  Stats_ = [SuccessEntry_,FailEntry,SuccessBio,FailBio] &

  SuccessEntry_ is SuccessEntry + 1.

addFailedEntryToStats(Stats,Stats_) :-
  Stats = [SuccessEntry,FailEntry,SuccessBio,FailBio] &
  Stats_ = [SuccessEntry,FailEntry_,SuccessBio,FailBio] &

  FailEntry_ is FailEntry + 1.

addSuccessfulBioCheckToStats(Stats,Stats_) :-
  Stats = [SuccessEntry,FailEntry,SuccessBio,FailBio] &
  Stats_ = [SuccessEntry,FailEntry,SuccessBio_,FailBio] &

  SuccessBio_ is SuccessBio + 1.

addFailedBioCheckToStats(Stats,Stats_) :-
  Stats = [SuccessEntry,FailEntry,SuccessBio,FailBio] &
  Stats_ = [SuccessEntry,FailEntry,SuccessBio,FailBio_] &

  FailBio_ is FailBio + 1.

unlockDoor(DoorLatchAlarm,Config,DoorLatchAlarm_) :-
  DoorLatchAlarm =
    [CurrentTime,CurrentDoor,_,_,_,_] &
  Config = [AlarmSilentDuration,LatchUnlockDuration,_,_,_,_,_,_] &
  DoorLatchAlarm_ = 
    [CurrentTime_,CurrentDoor_,CurrentLatch_,DoorAlarm_,LatchTimeout_,
     AlarmTimeout_] &

  0 =< CurrentTime &   % implicit precondition
  0 =< AlarmSilentDuration &   % implicit precondition
  0 =< LatchUnlockDuration &   % implicit precondition

  LatchTimeout_ is CurrentTime + LatchUnlockDuration &
  AlarmTimeout_ is CurrentTime + LatchUnlockDuration + AlarmSilentDuration &
  CurrentTime_ = CurrentTime &
  CurrentDoor_ = CurrentDoor &

  (LatchUnlockDuration neq 0 &   % implicit condition
   CurrentLatch_ = unlocked &
   DoorAlarm_ = silent
   or
   LatchUnlockDuration = 0 & 
   CurrentLatch_ = locked &
   (AlarmSilentDuration neq 0 &
    DoorAlarm_ = silent
    or
    AlarmSilentDuration = 0 &
    (CurrentDoor = open &
     DoorAlarm_ = alarming
     or
     CurrentDoor = closed &
     DoorAlarm_ = silent
    )
   )
  ).

lockDoor(DoorLatchAlarm,DoorLatchAlarm_) :-
  DoorLatchAlarm = [CurrentTime,CurrentDoor,_,_,_,_] &
  DoorLatchAlarm_ =
    [CurrentTime_,CurrentDoor_,CurrentLatch_,DoorAlarm_,LatchTimeout_,
     AlarmTimeout_] &

  CurrentLatch_ = locked &
  LatchTimeout_ = CurrentTime &
  AlarmTimeout_ = CurrentTime &
  CurrentTime_ = CurrentTime &
  CurrentDoor_ = CurrentDoor &

  (CurrentDoor = open &   % implicit condition
   DoorAlarm_ = alarming
   or
   CurrentDoor = closed &
   DoorAlarm_ = silent
  ).

certIssuerKnown(KeyStore,Certificate) :-
  KeyStore = [IssuerKey,_] &
  Certificate = [Id,_,_] & Id = [Issuer] &

  Issuer in D & dom(IssuerKey,D).

certOK(KeyStore,Certificate) :-
  KeyStore = [IssuerKey,_] &
  Certificate = [Id,_,IsValidatedBy] &
    Id = [Issuer] &

  certIssuerKnown(KeyStore,Certificate) &

  [Issuer,M] in IssuerKey & M in IsValidatedBy.

certIssuerIsThisTIS(KeyStore,Certificate) :-
  KeyStore = [_,OwnName] &
  Certificate = [Id,_,_] &
    Id = [Issuer] &

  OwnName neq {} &
  the(OwnName,Issuer).

authCertOK(KeyStore,Certificate) :-
  certIssuerIsThisTIS(KeyStore,Certificate) &
  certOK(KeyStore,Certificate).

newAuthCert(Token,KeyStore,Config,NewAuthCert,CurrentTime) :-
  Token = [TokenID,IDCert,PrivCert,_,_] &
    PrivCert = [_,_,_,_,_,R,C] &
    IDCert = [[I1,_,_],[_,_]] & 
  KeyStore = [IssuerKey,OwnName] &
  Config = [_,_,_,EnclaveClearance,AuthPeriod,_,_,_] &
  NewAuthCert = [Id,ValidityPeriod,IsValidatedBy,BaseCertId,T1,Role,Clearance] &
    Id = [Issuer] &

  validToken(Token) & 

  0 =< CurrentTime &   % implicit precondition

  OwnName neq {} &

  the(OwnName,Issuer) &
  [[R,CurrentTime],ValidityPeriod] in AuthPeriod &
  BaseCertId = I1 &
  T1 = TokenID &
  Role = R &
  minClearance(EnclaveClearance,C,Clearance) &
  IsValidatedBy = {M1} & [M2,M1] in IssuerKey & the(OwnName,M2).

% Why 
% currentUserToken \in \ran goodT 
% but 
% currentUserToken' \in \ran goodT 
% isn't when in both cases the inverse of goodT is taken?
addAuthCertToUserToken(UserToken,KeyStore,Config,CurrentTime,UserToken_) :-
  UserToken = [CurrentUserToken,UserTokenPresence] &
  UserToken_ = [CurrentUserToken_,UserTokenPresence_] &

  UserTokenPresence = present &
  CurrentUserToken = goodT(_) &

  validToken(T) & validToken(T_) &
      T = [TokenID,IDCert,PrivCert,IandACert,_] &
      T_ = [TokenID_,IDCert_,PrivCert_,IandACert_,AuthCert_] &
    goodT(T) = CurrentUserToken &
    goodT(T_) = CurrentUserToken_ &
    the(AuthCert_,NewAuthCert) &
    newAuthCert(T,KeyStore,Config,NewAuthCert,CurrentTime) &
    TokenID_ = TokenID &
    IDCert_ = IDCert &
    PrivCert_ = PrivCert &
    IandACert_ = IandACert &

  UserTokenPresence_ = UserTokenPresence.

% There's an implicit precondition on issuerCerts:
% the projection of the subject and subjectPubK
% fields must be a partial function
updateKeyStore(KeyStore,Enrol,KeyStore_) :-
  KeyStore = [IssuerKey,OwnName] &
  Enrol = [IdStationCert,IssuerCerts] &
    IdStationCert = [[_,_,_],[Subject,SubjectPubK]] &
  KeyStore_ = [IssuerKey_,OwnName_] &

  pfun(SubSubPK) & ran(IssuerCerts,SubSubPK) &     % implicit precondition

  ncomp({[Subject,Subject]},IssuerKeyR_,{}) &

  the(OwnName, M3) &
  the(OwnName_, Subject) &
  IssuerKey = {[M3,_]/IssuerKeyR} & comp({[M3,M3]},IssuerKeyR,{}) &
  IssuerKey_ = {[M3,SubjectPubK]/IssuerKeyR_} & comp({[M3,M3]},IssuerKeyR_,{}) &
  oplus(IssuerKeyR,SubSubPK,IssuerKeyR_) &

  delay(validEnrol(Enrol),false).

updateKeyStore_SAT(KeyStore,Enrol,KeyStore_) :-
  KeyStore = [IssuerKey,OwnName] &
  Enrol = [IdStationCert,IssuerCerts] &
    IdStationCert = [[_,_,_],[Subject,SubjectPubK]] &
  KeyStore_ = [IssuerKey_,OwnName_] &

  pfun(SubSubPK) & ran(IssuerCerts,SubSubPK) &     % implicit precondition

  ncomp({[Subject,Subject]},IssuerKeyR_,{}) &

  the(OwnName, M3) &
  the(OwnName_, Subject) &
  IssuerKey = {[M3,_]/IssuerKeyR} & comp({[M3,M3]},IssuerKeyR,{}) &
  IssuerKey_ = {[M3,SubjectPubK]/IssuerKeyR_} & comp({[M3,M3]},IssuerKeyR_,{}) &
  oplus(IssuerKeyR,SubSubPK,IssuerKeyR_) &

  validEnrol(Enrol).

updateKeyStoreFromFloppy(KeyStore,Floppy,KeyStore_) :-
  Floppy = [CurrentFloppy,_,_] &

  CurrentFloppy = enrolmentFile(Enrol) &
  updateKeyStore(KeyStore,Enrol,KeyStore_).

updateKeyStoreFromFloppy_SAT(KeyStore,Floppy,KeyStore_) :-
  Floppy = [CurrentFloppy,_,_] &

  CurrentFloppy = enrolmentFile(Enrol) &
  updateKeyStore_SAT(KeyStore,Enrol,KeyStore_).

adminLogon(Admin,AdminToken,Admin_) :-
  Admin = [RolePresent,_,_] &
  Admin_ = [RolePresent_,AvailableOps_,CurrentAdminOp_] &
  AdminToken = [CurrentAdminToken,_] &

  RolePresent = {} &

  validToken(T) & 
      T = [_,_,_,_,AuthCert] &
      the(AuthCert,AC) & AC = [_,_,_,_,_,Role,_] & 
    CurrentAdminToken = goodT(T) &
    TheRolePresent = Role & the(RolePresent_,TheRolePresent) & 
  
  CurrentAdminOp_ = {} &

  (Role = guard &   % implicit condition
   AvailableOps_ = {overrideLock}
   or
   Role = auditManager &
   AvailableOps_ = {archiveLog}
   or
   Role = securityOfficer &
   AvailableOps_ = {updateConfigData, shutdownOp}
  ).

adminLogout(Admin,Admin_) :-
  Admin = [RolePresent,_,_] &
  Admin_ = [RolePresent_,AvailableOps_,CurrentAdminOp_] &

  RolePresent neq {} &

  RolePresent_ = {} &
  CurrentAdminOp_ = {} &

  AvailableOps_ = {}.   % implicit condition

adminStartOp(Admin,Keyboard,Admin_) :-
  Admin = [RolePresent,AvailableOps,CurrentAdminOp] &
  Admin_ = [RolePresent_,AvailableOps_,CurrentAdminOp_] &
  Keyboard = [CurrentKeyedData,_] &

  RolePresent neq {} &
  CurrentAdminOp = {} &
  CurrentKeyedData = keyedOps(M1) & M1 in AvailableOps &

  RolePresent_ = RolePresent &
  M2 = M1 & the(CurrentAdminOp_,M2) &

  the(RolePresent,M3) &   % implicit condition
  (M3 = guard &
   AvailableOps_ = {overrideLock}
   or
   M3 = auditManager &
   AvailableOps_ = {archiveLog}
   or
   M3 = securityOfficer &
   AvailableOps_ = {updateConfigData, shutdownOp}
  ).

adminFinishOp(Admin,Admin_) :-
  Admin = [RolePresent,_,CurrentAdminOp] &
  Admin_ = [RolePresent_,AvailableOps_,CurrentAdminOp_] &

  RolePresent neq {} &
  CurrentAdminOp neq {} &

  RolePresent_ = RolePresent &
  CurrentAdminOp_ = {} &

  the(RolePresent,M3) &   % implicit condition
  (M3 = guard &
   AvailableOps_ = {overrideLock}
   or
   M3 = auditManager &
   AvailableOps_ = {archiveLog}
   or
   M3 = securityOfficer &
   AvailableOps_ = {updateConfigData, shutdownOp}
  ).

resetScreenMessage(_Internal,
                   _Admin,CurrentScreen,Internal_,Admin_,CurrentScreen_) :-
  CurrentScreen = [_,ScreenMsg,_] &
  Internal_ = [Status_,EnclaveStatus_,_] &
  Admin_ = [RolePresent_,_,_] &
  CurrentScreen_ = [_,ScreenMsg_,_] &

  (Status_ nin {quiescent, waitingRemoveTokenFail} &
   ScreenMsg_ = busy
   or
   Status_ in {quiescent, waitingRemoveTokenFail} &
   (EnclaveStatus_ = enclaveQuiescent & 
    RolePresent_ = {} &
    ScreenMsg_ = welcomeAdmin
    or
    EnclaveStatus_ = enclaveQuiescent & 
    RolePresent_ neq {} &
    ScreenMsg_ = requestAdminOp
    or
    EnclaveStatus_ = waitingRemoveAdminTokenFail & 
    ScreenMsg_ = removeAdminToken
    or
    EnclaveStatus_ nin {enclaveQuiescent, waitingRemoveAdminTokenFail} &
    ScreenMsg_ = ScreenMsg
   )
  ).

userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,Finger,_,Floppy,Keyboard,Config,_,KeyStore,Admin,_,Internal,
     _,CurrentScreen] &
    Internal = [Status,EnclaveStatus,TokenRemovalTimeout] &
  IDStation_ = 
    [_,AdminToken_,Finger_,_,Floppy_,
     Keyboard_,Config_,_,KeyStore_,Admin_,_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,TokenRemovalTimeout_] &

  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld = [TISControlledRealWorld,_] &
    RealWorld_ = [TISControlledRealWorld_,_] &

  Config_ = Config &
  AdminToken_ = AdminToken &
  KeyStore_ = KeyStore &
  Admin_ = Admin &
  Keyboard_ = Keyboard &
  Floppy_ = Floppy &
  Finger_ = Finger &

  TISControlledRealWorld_ = TISControlledRealWorld &

  resetScreenMessage(Internal,
                     Admin,CurrentScreen,Internal_,Admin_,CurrentScreen_) &

  EnclaveStatus_ = EnclaveStatus &
  (Status = waitingEntry &
   0 =< TokenRemovalTimeout_   % implicit condition
   or 
   TokenRemovalTimeout_ = TokenRemovalTimeout).

userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,
     _,_,Stats_,_,_,AuditLog_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  addFailedEntryToStats(Stats,Stats_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status in {gotUserToken,waitingUpdateToken,waitingFinger,gotFinger,
             waitingEntry} &
  UserTokenPresence = absent &

  CurrentDisplay_ = welcome &
  Status_ = quiescent &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

readUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    Internal = [Status,EnclaveStatus,_] &
    UserToken = [_,UserTokenPresence] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,
     AuditLog_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus in {enclaveQuiescent, waitingRemoveAdminTokenFail} &

  Status = quiescent &
  UserTokenPresence = present &
  CurrentDisplay_ = wait &
  Status_ = gotUserToken &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisReadUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readUserToken(IDStation,RealWorld,RealWorld_,IDStation_).

userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime) :-
  UserToken = [CurrentUserToken,_] &

  0 =< CurrentTime &   % implicit precondition

  CurrentUserToken = goodT(_) & 

  tokenWithValidAuth(T) &
    goodT(T) = CurrentUserToken & 
      T = [_,IDCert,_,_,AuthCert] &
        IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] & 

    CurrentTime in M1 & 
      the(AuthCert,M2) & M2 = [M3,M1,M4,_,_,_,_] &

    certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy]) &

    authCertOK(KeyStore,[M3,M1,M4]).

% In Z they existencially quantify CurrentToken but
% then they ask for \theta ValidToken.
% The quantification over CurrentToken is necessary
% to have access to variable now, declared in 
% RealWorldChanges which is a schema included in the schema 
% where userTokenOK is called.
% This complicates the negation of userTokenOK.
% Then, instead, we declare now as a variable in
% userTokenOK and so the quantification is done over
% a validToken, instead of on a currentToken
userTokenOK(KeyStore,UserToken,CurrentTime,Now) :-
  UserToken = [CurrentUserToken,_] &

  0 =< CurrentTime &   % implicit precondition

  CurrentUserToken = goodT(_) & 

  currentToken(Token,Now) & 
    Token = [_,IDCert,PrivCert,IandACert,_] &
      IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] & 
      PrivCert = [PCId,PCValidityPeriod,PCIsValidatedBy,_,_,_,_] &
      IandACert = [IAId,IAValidityPeriod,IAIsValidatedBy,_,_,_] &

    goodT(Token) = CurrentUserToken & 

    Now = CurrentTime &

    certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy]) &

    certOK(KeyStore,[PCId,PCValidityPeriod,PCIsValidatedBy]) &

    certOK(KeyStore,[IAId,IAValidityPeriod,IAIsValidatedBy]).

bioCheckNotRequired(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,Internal,
     _,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = gotUserToken &
  UserTokenPresence = present &

  userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime) &

  Status_ = waitingEntry &
  CurrentDisplay_ = wait &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

bioCheckRequired(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,Internal,
     _,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [Now,_,_,_,_,_,_] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = gotUserToken &
  UserTokenPresence = present &

  delay(not_userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime),false) &
    userTokenOK(KeyStore,UserToken,CurrentTime,Now) &

  CurrentDisplay_ = insertFinger &
  Status_ = waitingFinger &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

validateUserTokenOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  bioCheckRequired(IDStation,RealWorld,RealWorld_,IDStation_) 
  or
  bioCheckNotRequired(IDStation,RealWorld,RealWorld_,IDStation_).

validateUserTokenFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,Internal,
     _,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  RealWorld = [_,TISMonitoredRealWorld] &
    TISMonitoredRealWorld = [Now,_,_,_,_,_,_] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = gotUserToken &
  UserTokenPresence = present &

  delay(not_userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime) &
    not_userTokenOK(KeyStore,UserToken,CurrentTime,Now),false) &

  CurrentDisplay_ = removeToken &
  Status_ = waitingRemoveTokenFail &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisValidateUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateUserTokenOK(IDStation,RealWorld,RealWorld_,IDStation_) 
  or
  validateUserTokenFail(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = gotUserToken.

readFingerOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,Finger,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,
     Internal,_,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Finger = [_,FingerPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingFinger &
  FingerPresence = present &
  UserTokenPresence = present &

  userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime) &

  Status_ = gotFinger &
  CurrentDisplay_ = wait &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

noFinger(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [UserToken,_,Finger,_,_,_,_,_,_,_,_,Internal,_,_] &
    Internal = [Status,_,_] &
    Finger = [_,FingerPresence] &
    UserToken = [_,UserTokenPresence] &

  IDStation_ = IDStation &
  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld = [TISControlledRealWorld,_] &
    RealWorld_ = [TISControlledRealWorld_,_] &

  TISControlledRealWorld_ = TISControlledRealWorld &

  Status = waitingFinger &
  FingerPresence = absent &
  UserTokenPresence = present.

fingerTimeout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingFinger &
  UserTokenPresence = present &

  Status_ = waitingRemoveTokenFail &
  CurrentDisplay_ = removeToken &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisReadFinger(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readFingerOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  fingerTimeout(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  noFinger(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = waitingFinger.

fingerOK(Finger,_UserToken) :-
  Finger = [CurrentFinger,_] &

  CurrentFinger = goodFP(_).

validateFingerOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,Finger,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,
     Internal,_,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &

  addSuccessfulBioCheckToStats(Stats,Stats_) &
  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = gotFinger &
  UserTokenPresence = present &

  fingerOK(Finger,UserToken) &

  Status_ = waitingUpdateToken &
  CurrentDisplay_ = wait &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

validateFingerFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &

  addFailedBioCheckToStats(Stats,Stats_) &
  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = gotFinger &
  UserTokenPresence = present &

  Status_ = waitingRemoveTokenFail &
  CurrentDisplay_ = removeToken &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisValidateFinger(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateFingerOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  validateFingerFail(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = gotFinger.

writeUserTokenOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,Internal,
     _,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &
  addAuthCertToUserToken(UserToken,KeyStore,Config,CurrentTime,UserToken_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingUpdateToken &
  UserTokenPresence = present &

  Status_ = waitingEntry &
  CurrentDisplay_ = wait &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

writeUserTokenFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,_,AuditLog,Internal,
     _,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &
  addAuthCertToUserToken(UserToken,KeyStore,Config,CurrentTime,UserToken_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingEntry &
  UserTokenPresence = present &

  Status_ = waitingEntry &
  CurrentDisplay_ = tokenUpdateFailed &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

writeUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  writeUserTokenOK(IDStation,RealWorld,RealWorld_,IDStation_) 
  or 
  writeUserTokenFail(IDStation,RealWorld,RealWorld_,IDStation_).

tisWriteUserToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  writeUserToken(IDStation,RealWorld,RealWorld__,IDStation__) &
  updateUserToken(IDStation__,RealWorld__,RealWorld_,IDStation_)
  or
  userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = waitingUpdateToken.

% Why not to ask currentUserToken \in \ran goodT
% as in other similar operations?
userAllowedEntry(UserToken,Config,CurrentTime) :-
  UserToken = [CurrentUserToken,_] &
  Config = [_,_,_,_,_,EntryPeriod,_,_] &

  0 =< CurrentTime  &  % implicit precondition

  goodT(T) = CurrentUserToken &
    T = [_,_,PrivCert,_,AuthCert] & 
  (validToken(T) & 
     CurrentTime in M1 & [Role,[Class,M1]] in EntryPeriod &
       PrivCert = [_,_,_,_,_,Role,Clearance] & Clearance = [Class]
   or
   tokenWithValidAuth(T) & 
     the(AuthCert,AC) & 
       AC = [_,_,_,_,_,ACRole,ACClearance] & ACClearance = [ACClass] &
     CurrentTime in M2 & [ACRole,[ACClass,M2]] in EntryPeriod
  ).

entryOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Config = [_,_,TokenRemovalDuration,_,_,_,_,_] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,TokenRemovalTimeout_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingEntry &
  UserTokenPresence = present &

  userAllowedEntry(UserToken,Config,CurrentTime) &

  Status_ = waitingRemoveTokenSuccess &
  CurrentDisplay_ = openDoor &
  TokenRemovalTimeout_ is CurrentTime + TokenRemovalDuration &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

entryNotAllowed(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,TokenRemovalTimeout] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,TokenRemovalTimeout_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingEntry &
  UserTokenPresence = present &

  not_userAllowedEntry(UserToken,Config,CurrentTime) &

  Status_ = waitingRemoveTokenFail &
  CurrentDisplay_ = removeToken &
  TokenRemovalTimeout_ = TokenRemovalTimeout &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisValidateEntry(IDStation,RealWorld,RealWorld_,IDStation_) :-
  entryOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or 
  entryNotAllowed(IDStation,RealWorld,RealWorld_,IDStation_)
  or 
  userTokenTorn(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = waitingEntry.

unlockDoorOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation =
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &

  unlockDoor(DoorLatchAlarm,Config,DoorLatchAlarm_) &
  addSuccessfulEntryToStats(Stats,Stats_) &
  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingRemoveTokenSuccess &
  UserTokenPresence = absent &

  Status_ = quiescent &
  CurrentDisplay_ = doorUnlocked &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

waitingTokenRemoval(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [UserToken,_,_,DoorLatchAlarm,_,_,_,_,_,_,_,Internal,_,_] &
    Internal = [Status,_,TokenRemovalTimeout] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    UserToken = [_,UserTokenPresence] &

  IDStation_ = IDStation &
  realWorldChanges(RealWorld,RealWorld_) &
    RealWorld = [TISControlledRealWorld,_] &
    RealWorld_ = [TISControlledRealWorld_,_] &

  TISControlledRealWorld_ = TISControlledRealWorld &

  Status in {waitingRemoveTokenSuccess, waitingRemoveTokenFail} &
  (Status neq waitingRemoveTokenSuccess or CurrentTime =< TokenRemovalTimeout) &
  UserTokenPresence = present.

tokenRemovalTimeout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,
     Internal,_,CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [Status,_,TokenRemovalTimeout] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Stats_ = Stats &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingRemoveTokenSuccess &
  CurrentTime > TokenRemovalTimeout &
  UserTokenPresence = present &

  Status_ = waitingRemoveTokenFail &
  CurrentDisplay_ = removeToken &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisUnlockDoor(IDStation,RealWorld,RealWorld_,IDStation_) :-
  unlockDoorOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  waitingTokenRemoval(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = waitingRemoveTokenSuccess
  or
  tokenRemovalTimeout(IDStation,RealWorld,RealWorld_,IDStation_).

failedAccessTokenRemoved(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,_,_,Config,Stats,_,_,AuditLog,Internal,_,
     CurrentScreen] &
    UserToken = [_,UserTokenPresence] &
    Internal = [Status,_,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,_,_,DoorLatchAlarm_,_,_,_,Stats_,_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  userEntryContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  UserToken_ = UserToken &
  DoorLatchAlarm_ = DoorLatchAlarm &

  addFailedEntryToStats(Stats,Stats_) &
  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  Status = waitingRemoveTokenFail &
  UserTokenPresence = absent &

  Status_ = quiescent &
  CurrentDisplay_ = welcome &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisCompleteFailedAccess(IDStation,RealWorld,RealWorld_,IDStation_) :-
  failedAccessTokenRemoved(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  waitingTokenRemoval(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
      Internal = [Status,_,_] &
  Status = waitingRemoveTokenFail.

tisUserEntryOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  tisReadUserToken(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisValidateUserToken(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisReadFinger(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisValidateFinger(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisWriteUserToken(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisValidateEntry(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisUnlockDoor(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisCompleteFailedAccess(IDStation,RealWorld,RealWorld_,IDStation_).

enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,_,_,_,_,Stats,_,_,_,Internal,_,_] &
    Internal = [_,_,TokenRemovalTimeout] &
  RealWorld = [TISControlledRealWorld,_] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,_,_,_,_,Stats_,_,_,_,Internal_,_,_] &
    Internal_ = [_,_,TokenRemovalTimeout_] &
  RealWorld_ = [TISControlledRealWorld_,_] &

  realWorldChanges(RealWorld,RealWorld_) &

  TISControlledRealWorld_ = TISControlledRealWorld &

  UserToken_ = UserToken &
  AdminToken_ = AdminToken &
  Finger_ = Finger &
  Stats_ = Stats &

  TokenRemovalTimeout_ = TokenRemovalTimeout.
  
enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,DoorLatchAlarm,Floppy,Keyboard,Config,_,_,Admin,_,_,_,_] &
  IDStation_ =
    [_,_,_,DoorLatchAlarm_,Floppy_,Keyboard_,Config_,_,_,Admin_,_,_,_,_] &

  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Keyboard_ = Keyboard &
  Admin_ = Admin &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Config_ = Config &
  Floppy_ = Floppy.

requestEnrolment(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,AuditLog,Internal,_,
     CurrentScreen] &
    Floppy = [_,_,FloppyPresence] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,AuditLog_,Internal_,
    CurrentDisplay_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  KeyStore_ = KeyStore &
  AuditLog_ = AuditLog &
  Internal_ = Internal &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = notEnrolled &
  FloppyPresence = absent &

  ScreenMsg_ = insertEnrolmentData &
  CurrentDisplay_ = blank &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

readEnrolmentFloppy(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,_,Internal,_,CurrentScreen] &
    Floppy = [_,_,FloppyPresence] &
    Internal = [Status,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,_,Internal_,CurrentDisplay_,
     CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  KeyStore_ = KeyStore &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = notEnrolled &
  FloppyPresence = present &

  ScreenMsg_ = validatingEnrolmentData &

  EnclaveStatus_ = waitingEnrol &
  Status_ = Status &
  CurrentDisplay_ = blank &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

readEnrolmentData(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readEnrolmentFloppy(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  requestEnrolment(IDStation,RealWorld,RealWorld_,IDStation_).

enrolmentDataOK(Floppy,_KeyStore) :-
  Floppy = [CurrentFloppy,_,_] &

  CurrentFloppy = enrolmentFile(Enrol) &
    delay(validEnrol(Enrol),false).

validateEnrolmentDataOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,AuditLog,Internal,_,
     CurrentScreen] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  updateKeyStoreFromFloppy(KeyStore,Floppy,KeyStore_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingEnrol &

  ScreenMsg_ = welcomeAdmin &

  EnclaveStatus_ = enclaveQuiescent &
  Status_ = quiescent &
  CurrentDisplay_ = welcome &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig &    % implicit condition

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_) &
        enrolmentDataOK(Floppy,KeyStore),false).

validateEnrolmentDataOK_SAT(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,AuditLog,Internal,_,
     CurrentScreen] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  updateKeyStoreFromFloppy_SAT(KeyStore,Floppy,KeyStore_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingEnrol &

  ScreenMsg_ = welcomeAdmin &

  EnclaveStatus_ = enclaveQuiescent &
  Status_ = quiescent &
  CurrentDisplay_ = welcome &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig &    % implicit condition

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_) &
        enrolmentDataOK(Floppy,KeyStore),false).

validateEnrolmentDataOK_P4(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,AuditLog,Internal,_,
     CurrentScreen] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,AuditLog_,Internal_,
     CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  updateKeyStoreFromFloppy(KeyStore,Floppy,KeyStore_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingEnrol &

  ScreenMsg_ = welcomeAdmin &

  EnclaveStatus_ = enclaveQuiescent &
  Status_ = quiescent &
  CurrentDisplay_ = welcome &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig &    % implicit condition

  addElementsToLog(Config,_NewElements,AuditLog,AuditLog_) &
  delay(enrolmentDataOK(Floppy,KeyStore),false).


validateEnrolmentDataFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,Floppy,_,Config,Stats,KeyStore,_,AuditLog,Internal,_,
     CurrentScreen] &
    Internal = [Status,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,AuditLog_,Internal_,
    CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  KeyStore_ = KeyStore &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingEnrol &

  not_enrolmentDataOK(Floppy,KeyStore) & 

  ScreenMsg_ = enrolmentFailed &

  EnclaveStatus_ = waitingEndEnrol &
  Status_ = Status &
  CurrentDisplay_ = blank &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

validateEnrolmentData(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateEnrolmentDataOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  validateEnrolmentDataFail(IDStation,RealWorld,RealWorld_,IDStation_).

validateEnrolmentData_P4(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateEnrolmentDataOK_P4(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  validateEnrolmentDataFail(IDStation,RealWorld,RealWorld_,IDStation_).

failedEnrolFloppyRemoved(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,Stats,KeyStore,_,_,Internal,_,CurrentScreen] &
    Internal = [Status,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,_,_,_,_,Stats_,KeyStore_,_,_,Internal_,CurrentDisplay_,
     CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  KeyStore_ = KeyStore &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingEndEnrol &

  ScreenMsg_ = insertEnrolmentData &

  EnclaveStatus_ = notEnrolled &
  Status_ = Status &
  CurrentDisplay_ = blank &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

waitingFloppyRemoval(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,Floppy,_,_,_,_,_,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    Floppy = [_,_,FloppyPresence] &

  enrolContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  IDStation_ = IDStation &

  EnclaveStatus = waitingEndEnrol &
  FloppyPresence = present.

completeFailedEnrolment(IDStation,RealWorld,RealWorld_,IDStation_) :-
  failedEnrolFloppyRemoved(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  waitingFloppyRemoval(IDStation,RealWorld,RealWorld_,IDStation_).

tisEnrolOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readEnrolmentData(IDStation,RealWorld,RealWorld_,IDStation_)
  or 
  validateEnrolmentData(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  completeFailedEnrolment(IDStation,RealWorld,RealWorld_,IDStation_).

tisEnrolOp_P4(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readEnrolmentData(IDStation,RealWorld,RealWorld_,IDStation_)
  or 
  validateEnrolmentData_P4(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  completeFailedEnrolment(IDStation,RealWorld,RealWorld_,IDStation_).

adminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,DoorLatchAlarm,Floppy,
     Keyboard,Config,_,KeyStore,Admin,_,Internal,CurrentDisplay,CurrentScreen] &
    AdminToken = [_,AdminTokenPresence] &
    Internal = [Status,_,_] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,Keyboard_,Config_,_,KeyStore_,
     Admin_,_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &

  Config_ = Config &
  Floppy_ = Floppy &
  Keyboard_ = Keyboard &
  DoorLatchAlarm_ = DoorLatchAlarm &
  KeyStore_ = KeyStore &
  
  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  resetScreenMessage(Internal,
                     Admin,CurrentScreen,Internal_,Admin_,CurrentScreen_) &

  AdminTokenPresence = absent &

  Status_ = Status &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = enclaveQuiescent.

badAdminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,AuditLog,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
  IDStation_ = [_,_,_,_,_,_,Config,_,_,_,AuditLog_,_,_,_] &

  adminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  EnclaveStatus in {gotAdminToken, waitingStartAdminOp, waitingFinishAdminOp}.

badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,Stats,_,Admin,_,Internal,_,CurrentScreen] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = [_,_,_,_,_,_,_,Stats_,_,Admin_,_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  badAdminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) &
  adminLogout(Admin,Admin_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus in {waitingStartAdminOp, waitingFinishAdminOp} &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

loginAborted(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,Stats,_,Admin,_,Internal,_,CurrentScreen] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = [_,_,_,_,_,_,Config_,Stats_,_,Admin_,_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  badAdminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) &
  Admin_ = Admin &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = gotAdminToken &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = displayConfigData(Config_).    % implicit condition

loginContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,
     Keyboard,Config,_,KeyStore,_,_,Internal,CurrentDisplay,_] &
    Internal = [Status,_,_] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,_,KeyStore_,_,_,Internal_,CurrentDisplay_,_] &
    Internal_ = [Status_,_,_] &

  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Keyboard_ = Keyboard &
  KeyStore_ = KeyStore &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Config_ = Config &
  Floppy_ = Floppy &

  Status_ = Status &
  CurrentDisplay_ = CurrentDisplay.

readAdminToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,_,_,_,Config,_,_,Admin,AuditLog,Internal,_,CurrentScreen] &
    Internal = [Status,EnclaveStatus,_] &
    Admin = [RolePresent,_,_] &
    AdminToken = [_,AdminTokenPresence] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &

  loginContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Admin_ = Admin &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  Status in {quiescent, waitingRemoveTokenFail} &

  EnclaveStatus = enclaveQuiescent &
  RolePresent = {} &
  AdminTokenPresence = present &

  EnclaveStatus_ = gotAdminToken &
  CurrentScreen_ = CurrentScreen.

tisReadAdminToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  readAdminToken(IDStation,RealWorld,RealWorld_,IDStation_).

adminTokenOK(AdminToken,KeyStore,CurrentTime) :-
  AdminToken = [CurrentAdminToken,_] &

  0 =< CurrentTime &   % implicit precondition

  CurrentAdminToken = goodT(_) & 

  tokenWithValidAuth(T) &
    goodT(T) = CurrentAdminToken & 
      T = [_,IDCert,_,_,AuthCert] &
        IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] &

     certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy]) &

    the(AuthCert,AC) & 
      AC = [ACId,ACValidityPeriod,ACIsValidatedBy,_,_,Role,_] & 
    authCertOK(KeyStore,[ACId,ACValidityPeriod,ACIsValidatedBy]) &

    Role in ADMINPRIVILEGE & adminprivilege(ADMINPRIVILEGE) &

    CurrentTime in ACValidityPeriod.

validateAdminTokenOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,DoorLatchAlarm,_,
     _,Config,Stats,KeyStore,Admin,AuditLog,Internal,_,CurrentScreen] &
    AdminToken = [_,AdminTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  loginContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  adminLogon(Admin,AdminToken,Admin_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = gotAdminToken &
  AdminTokenPresence = present &

  adminTokenOK(AdminToken,KeyStore,CurrentTime) &

  ScreenMsg_ = requestAdminOp &

  EnclaveStatus_ = enclaveQuiescent &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

validateAdminTokenFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,DoorLatchAlarm,_,
     _,Config,Stats,KeyStore,Admin,AuditLog,Internal,_,CurrentScreen] &
    AdminToken = [_,AdminTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  loginContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Admin_ = Admin &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = gotAdminToken &
  AdminTokenPresence = present &

  not_adminTokenOK(AdminToken,KeyStore,CurrentTime) &

  ScreenMsg_ = removeAdminToken &

  EnclaveStatus_ = waitingRemoveAdminTokenFail &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisValidateAdminToken(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateAdminTokenOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  validateAdminTokenFail(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  loginAborted(IDStation,RealWorld,RealWorld_,IDStation_).

failedAdminTokenRemoved(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,_,_,_,Config,Stats,_,Admin,AuditLog,Internal,CurrentDisplay,
     CurrentScreen] &
    AdminToken = [_,AdminTokenPresence] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
   [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,CurrentDisplay_,
    CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  loginContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Admin_ = Admin &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingRemoveAdminTokenFail &
  AdminTokenPresence = absent &

  ScreenMsg_ = welcomeAdmin &

  EnclaveStatus_ = enclaveQuiescent &

  CurrentDisplay_ = CurrentDisplay &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

waitingAdminTokenRemoval(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,AdminToken,_,_,_,_,_,_,_,_,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    AdminToken = [_,AdminTokenPresence] &

  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  IDStation_ = IDStation &

  EnclaveStatus = waitingRemoveAdminTokenFail &
  AdminTokenPresence = present.
  
tisCompleteFailedAdminLogon(IDStation,RealWorld,RealWorld_,IDStation_) :-
  failedAdminTokenRemoved(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  waitingAdminTokenRemoval(IDStation,RealWorld,RealWorld_,IDStation_).

tisAdminLogon(IDStation,RealWorld,RealWorld_,IDStation_) :-
  tisReadAdminToken(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisValidateAdminToken(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisCompleteFailedAdminLogon(IDStation,RealWorld,RealWorld_,IDStation_).

tokenRemovedAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation =
    [_,_,_,_,_,_,Config,Stats,_,Admin,AuditLog,Internal,_,CurrentScreen] &
    Admin = [RolePresent,_,_] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = [_,_,_,_,_,_,Config_,_,_,Admin_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  adminTokenTear(IDStation,RealWorld,RealWorld_,IDStation_) &
  adminLogout(Admin,Admin_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = enclaveQuiescent &
  RolePresent neq {} &

  ScreenStats_ = ScreenStats &    % implicit condition
  ScreenConfig_ = displayConfigData(Config_).    % implicit condition

adminTokenTimeout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,DoorLatchAlarm,_,
     _,Config,Stats,KeyStore,Admin,AuditLog,Internal,_,CurrentScreen] &
    AdminToken = [_,AdminTokenPresence] &
    DoorLatchAlarm = [CurrentTime,_,_,_,_,_] &
    Admin = [RolePresent,_,_] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  loginContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  adminLogout(Admin,Admin_) &
  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &
  resetScreenMessage(Internal,
                     Admin,CurrentScreen,Internal_,Admin_,CurrentScreen_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = enclaveQuiescent &
  AdminTokenPresence = present &
  RolePresent neq {} &

  not_adminTokenOK(AdminToken,KeyStore,CurrentTime) &

  EnclaveStatus_ = waitingRemoveAdminTokenFail &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisCompleteTimeoutAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  tisCompleteFailedAdminLogon(IDStation,RealWorld,RealWorld_,IDStation_).

tisAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) :-
  tokenRemovedAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  adminTokenTimeout(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisCompleteTimeoutAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_).

adminOpContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,Keyboard,_,_,KeyStore,_,_,_,_,_] &
  IDStation_ = [_,_,_,_,_,Keyboard_,_,_,KeyStore_,_,_,_,_,_] &

  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Keyboard_ = Keyboard &
  KeyStore_ = KeyStore.

adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,AdminToken,_,_,_,_,_,_,_,_,_,Internal,_,_] &
    AdminToken = [_,AdminTokenPresence] &
    Internal = [Status,EnclaveStatus,_] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,_,_,Internal_,_,_] &
    Internal_ = [Status_,_,_] &

  adminOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  EnclaveStatus = waitingStartAdminOp &
  AdminTokenPresence = present &

  Status_ = Status.

adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,AdminToken,_,_,_,_,_,_,_,Admin,_,Internal,CurrentDisplay,_] &
    Internal = [Status,EnclaveStatus,_] &
    AdminToken = [_,AdminTokenPresence] &
  IDStation_ = [_,_,_,_,_,_,_,_,_,Admin_,_,Internal_,CurrentDisplay_,_] &
    Internal_ = [Status_,EnclaveStatus_,_] &

  adminOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  adminFinishOp(Admin,Admin_) &

  EnclaveStatus = waitingFinishAdminOp &
  AdminTokenPresence = present &

  Status_ = Status &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = enclaveQuiescent.

startOpContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,AdminToken,_,DoorLatchAlarm,Floppy,
     Keyboard,Config,_,KeyStore,Admin,_,Internal,CurrentDisplay,_] &
    Internal = [Status,EnclaveStatus,_] &
    AdminToken = [_,AdminTokenPresence] &
    Admin = [RolePresent,_,_] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,_,KeyStore_,_,_,Internal_,CurrentDisplay_,_] &
    Internal_ = [Status_,_,_] &

  enclaveContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Keyboard_ = Keyboard &
  KeyStore_ = KeyStore &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Config_ = Config &
  Floppy_ = Floppy &

  EnclaveStatus = enclaveQuiescent &
  AdminTokenPresence = present &
  RolePresent neq nil &
  Status in {quiescent, waitingRemoveTokenFail} &

  Status_ = Status &
  CurrentDisplay_ = CurrentDisplay.

validateOpRequestOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation =
    [_,_,_,_,_,Keyboard,Config,Stats,_,Admin,AuditLog,_,_,CurrentScreen] &
    Keyboard = [CurrentKeyedData,KeyedDataPresence] &
    Admin = [_,AvailableOps,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  startOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  adminStartOp(Admin,Keyboard,Admin_) &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  KeyedDataPresence = present &
  CurrentKeyedData = keyedOps(M1) & M1 in AvailableOps &

  ScreenMsg_ = doingOp &

  EnclaveStatus_ = waitingStartAdminOp &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

validateOpRequestFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,_,_,Keyboard,Config,Stats,_,Admin,AuditLog,Internal,_,CurrentScreen] &
    Keyboard = [CurrentKeyedData,KeyedDataPresence] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ =
    [_,_,_,_,_,_,_,Stats_,_,Admin_,AuditLog_,Internal_,_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  startOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Admin_ = Admin &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  KeyedDataPresence = present &
  CurrentKeyedData in {noKB,badKB} &

  ScreenMsg_ = invalidRequest &

  EnclaveStatus_ = EnclaveStatus &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

noOpRequest(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,Keyboard,_,_,_,_,_,_,_,_] &
    Keyboard = [_,KeyedDataPresence] &

  IDStation_ = IDStation &

  startOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  KeyedDataPresence = absent.

validateOpRequest(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateOpRequestOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  validateOpRequestFail(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  noOpRequest(IDStation,RealWorld,RealWorld_,IDStation_).

tisStartAdminOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  validateOpRequest(IDStation,RealWorld,RealWorld_,IDStation_).

startArchiveLogOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,
     CurrentDisplay,CurrentScreen] &
    Floppy = [CurrentFloppy,_,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,AuditLog_,Internal_,CurrentDisplay_,
     CurrentScreen_] &
    Floppy_ = [CurrentFloppy_,WrittenFloppy_,FloppyPresence_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Admin_ = Admin &
  DoorLatchAlarm_ = DoorLatchAlarm &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,archiveLog) &
  FloppyPresence = present &

  FloppyPresence_ = FloppyPresence &
  CurrentFloppy_ = CurrentFloppy &

  ScreenMsg_ = doingOp &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = waitingFinishAdminOp &
  archiveLog(Config,AuditLog,Archive,AuditLog_) & 
    WrittenFloppy_ = auditFile(Archive) &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

startArchiveLogWaitingFloppy(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,_,Internal,
     CurrentDisplay,CurrentScreen] &
    Floppy = [CurrentFloppy,_,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Floppy_ = [CurrentFloppy_,_,FloppyPresence_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Admin_ = Admin &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Floppy_ = Floppy &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,archiveLog) &
  FloppyPresence = absent &

  FloppyPresence_ = FloppyPresence &
  CurrentFloppy_ = CurrentFloppy &

  ScreenMsg_ = insertBlankFloppy &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = EnclaveStatus &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

startArchiveLog(IDStation,RealWorld,RealWorld_,IDStation_) :-
  startArchiveLogOK(IDStation,RealWorld,RealWorld__,IDStation__) &
  updateFloppy(IDStation__,RealWorld__,IDStation_,RealWorld_)
  or
  startArchiveLogWaitingFloppy(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) &
      IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
        Internal = [_,EnclaveStatus,_] &
        Admin = [_,_,CurrentAdminOp] &
    EnclaveStatus = waitingStartAdminOp &
    the(CurrentAdminOp,archiveLog).

% Spivey p. 78: For the composition S;T to be defined, for
% each word x such that x' is a component of S and x itself
% is a component of T, the types of these two components
% must be the same.
% Spivey p. 134: If Op1 and Op2 are schemas describing two 
% operations, then Op1;Op2 is a schema which describes their
% sequential composition. For it to be defined, each
% dashed component of Op1 must match in type any undashed
% component of Op2 that matches it in name, and any other
% components, including inputs, outputs, and unmatched state
% variables, shared by Op1 and Op2 must have the same types
% in both of them. The components of Op1 Op2 are the merged
% components of Op1 and Op2, with the matching state variables
% hidden.
clearLogThenAddElements(Config,AuditLog,NewElements,Archive,AuditLog_) :-
  delay(clearLog(Config,AuditLog,Archive,AuditLog__),false) &
  delay(addElementsToLog(Config,NewElements,AuditLog__,AuditLog_),false).

finishArchiveLogOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Floppy = [CurrentFloppy,WrittenFloppy,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &

  the(CurrentAdminOp,archiveLog) &
  FloppyPresence = present &

  WrittenFloppy = CurrentFloppy &

  clearLogThenAddElements(Config,AuditLog,_NewElements,Archive,AuditLog_) & 
    WrittenFloppy = auditFile(Archive) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  ScreenMsg_ = requestAdminOp &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

finishArchiveLogNoFloppy(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Floppy = [_,_,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,archiveLog) &
  FloppyPresence = absent &

  ScreenMsg_ = archiveFailed &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

finishArchiveLogBadMatch(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Floppy = [CurrentFloppy,WrittenFloppy,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,archiveLog) &
  FloppyPresence = present &

  WrittenFloppy neq CurrentFloppy &

  ScreenMsg_ = archiveFailed &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

finishArchiveLogFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  finishArchiveLogBadMatch(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  finishArchiveLogNoFloppy(IDStation,RealWorld,RealWorld_,IDStation_).

finishArchiveLog(IDStation,RealWorld,RealWorld_,IDStation_) :-
  finishArchiveLogOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  finishArchiveLogFail(IDStation,RealWorld,RealWorld_,IDStation_) 
  or
  badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) &
      IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
        Internal = [_,EnclaveStatus,_] &
        Admin = [_,_,CurrentAdminOp] &
    EnclaveStatus = waitingFinishAdminOp &
    the(CurrentAdminOp,archiveLog).

tisArchiveLogOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  startArchiveLog(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  finishArchiveLog(IDStation,RealWorld,RealWorld_,IDStation_).

startUpdateConfigOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,_,_,CurrentDisplay,
     CurrentScreen] &
    Floppy = [_,_,FloppyPresence] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Floppy_ = Floppy &
  Config_ = Config &
  Admin_ = Admin &
  DoorLatchAlarm_ = DoorLatchAlarm &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,updateConfigData) &
  FloppyPresence = present &

  ScreenMsg_ = doingOp &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = waitingFinishAdminOp &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

startUpdateConfigWaitingFloppy(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,_,Internal,
     CurrentDisplay,CurrentScreen] &
    Admin = [_,_,CurrentAdminOp] &
    Floppy = [_,_,FloppyPresence] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Floppy_ = Floppy &
  Config_ = Config &
  Admin_ = Admin &
  DoorLatchAlarm_ = DoorLatchAlarm &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,updateConfigData) &
  FloppyPresence = absent &

  ScreenMsg_ = insertConfigData &
  CurrentDisplay_ = CurrentDisplay &

  EnclaveStatus_ = EnclaveStatus &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

startUpdateConfigData(IDStation,RealWorld,RealWorld_,IDStation_) :-
  startUpdateConfigOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  startUpdateConfigWaitingFloppy(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) &
      IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
        Internal = [_,EnclaveStatus,_] &
        Admin = [_,_,CurrentAdminOp] &
    EnclaveStatus = waitingStartAdminOp &
    the(CurrentAdminOp,updateConfigData).

% currentFloppy \in \ran configFile unnecessary because
% CurrentFloppy = configFile(Config_) won't unify
finishUpdateConfigDataOK(M,IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Floppy = [CurrentFloppy,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,_,AuditLog_,_,_,CurrentScreen_] &
  Config_ = [_,_,_,_,_,_,MinPreservedLogSize_,AlarmThresholdSize_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,updateConfigData) &

  CurrentFloppy = configFile(Config_) &

  ScreenMsg_ = requestAdminOp &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = displayConfigData(Config_) &    % implicit condition
  AlarmThresholdSize_ < MinPreservedLogSize_ &    % implicit condition
  MinPreservedLogSize_ =< M & maxSupportedLogSize(M).    % implicit condition

% currentFloppy \in \ran configFile negated by
% asserting that CurrentFloppy has one of the other
% forms of floppy_type
finishUpdateConfigDataFail(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Floppy = [CurrentFloppy,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,_,AuditLog_,_,_,CurrentScreen_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpFinishContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,updateConfigData) &

  (CurrentFloppy in {noFloppy, emptyFloppy, badFloppy}
   or
   CurrentFloppy = enrolmentFile(_)
   or
   CurrentFloppy = auditFile(_)
  ) &

  ScreenMsg_ = invalidData &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

finishUpdateConfigData(M,IDStation,RealWorld,RealWorld_,IDStation_) :-
  finishUpdateConfigDataOK(M,IDStation,RealWorld,RealWorld_,IDStation_)
  or
  finishUpdateConfigDataFail(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) &
      IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
        Internal = [_,EnclaveStatus,_] &
        Admin = [_,_,CurrentAdminOp] &
    EnclaveStatus = waitingFinishAdminOp &
    the(CurrentAdminOp,updateConfigData).

tisUpdateConfigDataOp(M,IDStation,RealWorld,RealWorld_,IDStation_) :-
  startUpdateConfigData(IDStation,RealWorld,RealWorld_,IDStation_) 
  or
  finishUpdateConfigData(M,IDStation,RealWorld,RealWorld_,IDStation_).

shutdownOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,
     Internal,_,CurrentScreen] &
    UserToken = [CurrentUserToken,_] &
    DoorLatchAlarm = [_,CurrentDoor,_,_,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,AuditLog_,Internal_,CurrentDisplay_,
     CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &
  lockDoor(DoorLatchAlarm,DoorLatchAlarm_) &
  adminLogout(Admin,Admin_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingStartAdminOp &
  the(CurrentAdminOp,shutdownOp) &
  CurrentDoor = closed &

  ScreenMsg_ = clear &

  EnclaveStatus_ = shutdown &
  CurrentDisplay_ = blank &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig &    % implicit condition
  % implicit condition
  (Status_ nin {gotFinger, waitingFinger, waitingUpdateToken, waitingEntry}
   or 
   validToken(T) & goodT(T) = CurrentUserToken
   or
   tokenWithValidAuth(T) & goodT(T) = CurrentUserToken
  ).

shutdownWaitingDoor(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,_,Internal,
     CurrentDisplay,CurrentScreen] &
    UserToken = [CurrentUserToken,_] &
    DoorLatchAlarm = [_,CurrentDoor,_,_,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    Internal = [_,EnclaveStatus,_] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,_,Internal_,CurrentDisplay_,CurrentScreen_] &
    Internal_ = [Status_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Config_ = Config &
  Floppy_ = Floppy &
  DoorLatchAlarm_ = DoorLatchAlarm &
  Admin_ = Admin &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  EnclaveStatus = waitingStartAdminOp &
  the(CurrentAdminOp,shutdownOp) &
  CurrentDoor = open &

  ScreenMsg_ = closeDoor &

  EnclaveStatus_ = EnclaveStatus &
  CurrentDisplay_ = CurrentDisplay &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig &    % implicit condition
  % implicit condition
  (Status_ nin {gotFinger, waitingFinger, waitingUpdateToken, waitingEntry}
   or 
   validToken(T) & goodT(T) = CurrentUserToken
   or
   tokenWithValidAuth(T) & goodT(T) = CurrentUserToken
  ).

tisShutdownOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  shutdownOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  shutdownWaitingDoor(IDStation,RealWorld,RealWorld_,IDStation_).

overrideDoorLockOK(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,Floppy,_,Config,Stats,_,Admin,AuditLog,_,_,
     CurrentScreen] &
    Admin = [_,_,CurrentAdminOp] &
    CurrentScreen = [ScreenStats,_,ScreenConfig] &
  IDStation_ = 
    [_,_,_,DoorLatchAlarm_,Floppy_,
     _,Config_,Stats_,_,Admin_,AuditLog_,Internal_,CurrentDisplay_,
     CurrentScreen_] &
    Internal_ = [_,EnclaveStatus_,_] &
    CurrentScreen_ = [ScreenStats_,ScreenMsg_,ScreenConfig_] &

  adminOpStartedContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  Floppy_ = Floppy &
  Config_ = Config &

  delay(addElementsToLog(Config,_NewElements,AuditLog,AuditLog_),false) &
  adminFinishOp(Admin,Admin_) &
  unlockDoor(DoorLatchAlarm,Config,DoorLatchAlarm_) &

  ScreenStats = displayStats(Stats) &    % implicit precondition
  ScreenConfig = displayConfigData(Config) &    % implicit precondition

  the(CurrentAdminOp,overrideLock) &

  ScreenMsg_ = requestAdminOp &
  CurrentDisplay_ = doorUnlocked &

  EnclaveStatus_ = enclaveQuiescent &

  ScreenStats_ = displayStats(Stats_) &    % implicit condition
  ScreenConfig_ = ScreenConfig.    % implicit condition

tisOverrideDoorLockOp(IDStation,RealWorld,RealWorld_,IDStation_) :-
  overrideDoorLockOK(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  badAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_) &
    IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
      Internal = [_,EnclaveStatus,_] &
      Admin = [_,_,CurrentAdminOp] &
  EnclaveStatus = waitingStartAdminOp &
  the(CurrentAdminOp,overrideLock).

initDoorLatchAlarm(DoorLatchAlarm) :-
  DoorLatchAlarm = [CurrentTime,CurrentDoor,_,_,LatchTimeout,AlarmTimeout] &

  zeroTime(T) &
  CurrentTime = T &
  CurrentDoor = closed &
  LatchTimeout = T &
  AlarmTimeout = T.

initKeyStore(KeyStore) :-
  KeyStore = [IssuerKey,OwnName] &

  IssuerKey = {} &
  OwnName = {}.

% Can't define AuthPeriod and EntryPeriod as in
% the specification due to the following reasons:
% a) {log} doesn't support infinite sets such as TIME
% b) integer intervals such as t \upto t+72000
%    aren't fully supported yet.
% c) relational constraints don't support 
%    RIS yet
% Then, I've approximated it as much as possible
initConfig(Config) :-
  Config = [AlarmSilentDuration,LatchUnlockDuration,TokenRemovalDuration,
            EnclaveClearance,AuthPeriod,EntryPeriod,_,_] &
    EnclaveClearance = [Class] &

  AlarmSilentDuration = 10 &
  LatchUnlockDuration = 150 &
  TokenRemovalDuration = 100 &
  Class = unmarked &
  AuthPeriod = 
    {[userOnly,M1], [guard,M1], [securityOfficer,M1], [auditManager,M1]} &
    foreach([T,TS] in M1, 
      0 =< T & 
      foreach(X in TS, 0 =< X & X =< 72000)) &
  EntryPeriod = cp(PRIVILEGE,cp(CLASS,M2)) &
    privilege(PRIVILEGE) & class(CLASS) &
    foreach(T in M2, 0 =< T).

initAdmin(Admin) :-
  Admin = [RolePresent,_,CurrentAdminOp] &

  RolePresent = {} &
  CurrentAdminOp = {}.

initStats(Stats) :-
  Stats = [SuccessEntry,FailEntry,SuccessBio,FailBio] &

  SuccessEntry = 0 &
  FailEntry = 0 &
  SuccessBio = 0 &
  FailBio = 0.

initAuditLog(AuditLog) :-
  AuditLog = [AauditLog,AuditAlarm] &

  AauditLog = {} &
  AuditAlarm = silent.

initIDStation(IDStation) :-
  IDStation = 
    [_,_,_,DoorLatchAlarm,_,_,Config,Stats,KeyStore,Admin,AuditLog,Internal,
     CurrentDisplay,CurrentScreen] &
    CurrentScreen = [_,ScreenMsg,_] &
    Internal = [Status,EnclaveStatus,_] &

  initDoorLatchAlarm(DoorLatchAlarm) &
  initConfig(Config) &
  initKeyStore(KeyStore) &
  initStats(Stats) &
  initAuditLog(AuditLog) &
  initAdmin(Admin) &

  ScreenMsg = clear &
  CurrentDisplay = blank &
  EnclaveStatus = notEnrolled &
  Status = quiescent.

startContext(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = 
    [UserToken,AdminToken,Finger,_,Floppy,Keyboard,Config,_,KeyStore,_,_,_,_,
     CurrentScreen] &
    CurrentScreen = [_,_,ScreenConfig] &
  IDStation_ = 
    [UserToken_,AdminToken_,Finger_,DoorLatchAlarm_,Floppy_,
     Keyboard_,Config_,Stats_,KeyStore_,Admin_,_,Internal_,_,CurrentScreen_] &
    DoorLatchAlarm_ = [_,_,CurrentLatch_,DoorAlarm_,_,_] &
    Admin_ = [_,AvailableOps_,_] &
    Internal_ = [_,_,TokenRemovalTimeout_] &
    CurrentScreen_ = [ScreenStats_,_,ScreenConfig_] &

  realWorldChanges(RealWorld,RealWorld_) &

  Config_ = Config &
  KeyStore_ = KeyStore &

  initDoorLatchAlarm(DoorLatchAlarm_) &

  CurrentLatch_ = locked &   % implict postcondition
  DoorAlarm_ = silent &   % implict postcondition

  initStats(Stats_) &
  initAdmin(Admin_) &

  AvailableOps_ = {} &   % implict postcondition

  UserToken_ = UserToken &
  AdminToken_ = AdminToken &
  Finger_ = Finger &
  Floppy_ = Floppy &
  Keyboard_ = Keyboard &

  ScreenStats_ = displayStats(Stats_) &
  ScreenConfig_ = ScreenConfig &   % implicit condition
  0 =< TokenRemovalTimeout_.   % implicit condition

% This operation calls addElementsToLog
% having existensially quantified a variable with the name
% newElements. With our encoding of addElementsToLog we
% simply pass in the quatified variable as the second
% argument to addElementsToLog
startNonEnrolledStation(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,_,KeyStore,_,AuditLog,_,_,_] &
    KeyStore = [_,OwnName] &
  IDStation_ = 
    [_,_,_,_,_,_,_,_,_,_,AuditLog_,Internal_,CurrentDisplay_,CurrentScreen_] &
    CurrentScreen_ = [_,ScreenMsg_,_] &
    Internal_ = [Status_,EnclaveStatus_,_] &

  startContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  OwnName = {} &

  ScreenMsg_ = insertEnrolmentData &

  CurrentDisplay_ = blank &
  EnclaveStatus_ = notEnrolled &
  Status_ = quiescent &

  delay(addElementsToLog(Config,NewElements,AuditLog,AuditLog_),false) &
    _StartUnenrolledTISElement in NewElements.

startEnrolledStation(IDStation,RealWorld,RealWorld_,IDStation_) :-
  IDStation = [_,_,_,_,_,_,Config,_,KeyStore,_,AuditLog,_,_,_] &
    KeyStore = [_,OwnName] &
  IDStation_ = 
    [_,_,_,_,_,_,_,_,_,_,AuditLog_,Internal_,CurrentDisplay_,CurrentScreen_] &
    CurrentScreen_ = [_,ScreenMsg_,_] &
    Internal_ = [Status_,EnclaveStatus_,_] &

  startContext(IDStation,RealWorld,RealWorld_,IDStation_) &

  OwnName neq {} &

  ScreenMsg_ = welcomeAdmin &

  CurrentDisplay_ = welcome &
  EnclaveStatus_ = enclaveQuiescent &
  Status_ = quiescent &

  delay(addElementsToLog(Config,NewElements,AuditLog,AuditLog_),false) &
    _StartEnrolledTISElement in NewElements.

tisStartup(IDStation,RealWorld,RealWorld_,IDStation_) :-
  startEnrolledStation(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  startNonEnrolledStation(IDStation,RealWorld,RealWorld_,IDStation_).

tisIdle(IDStation,TISControlledRealWorld,TISControlledRealWorld_,IDStation_) :-
  IDStation = [UserToken,AdminToken,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Internal = [Status,EnclaveStatus,_] &
    UserToken = [_,UserTokenPresence] &
    AdminToken = [_,AdminTokenPresence] &
    Admin = [RolePresent,_,_] &

  IDStation_ = IDStation &
  TISControlledRealWorld_ = TISControlledRealWorld &

  Status = quiescent &
  EnclaveStatus = enclaveQuiescent &
  UserTokenPresence = absent &
  AdminTokenPresence = absent &
  RolePresent = {}.

tisAdminOp(M,IDStation,RealWorld,RealWorld_,IDStation_) :-
  tisOverrideDoorLockOp(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisShutdownOp(IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisUpdateConfigDataOp(M,IDStation,RealWorld,RealWorld_,IDStation_)
  or
  tisArchiveLogOp(IDStation,RealWorld,RealWorld_,IDStation_).

tisProcessing(M,IDStation,RealWorld,RealWorld_,IDStation_) :-
  (tisEnrolOp(IDStation,RealWorld,RealWorld_,IDStation_)
   or
   tisUserEntryOp(IDStation,RealWorld,RealWorld_,IDStation_)
   or
   tisAdminLogon(IDStation,RealWorld,RealWorld_,IDStation_)
   or
   tisStartAdminOp(IDStation,RealWorld,RealWorld_,IDStation_)
   or
   tisAdminOp(M,IDStation,RealWorld,RealWorld_,IDStation_)
   or
   tisAdminLogout(IDStation,RealWorld,RealWorld_,IDStation_)
   or
   RealWorld = [TISControlledRealWorld,_] &
   RealWorld_ = [TISControlledRealWorld_,_] &
   tisIdle(IDStation,
           TISControlledRealWorld,TISControlledRealWorld_,IDStation_)
  ) & delay(logChange(IDStation,IDStation_),false).

% negations

%  AuthCert neq {} unnecessary, the(AuthCert,TheAuthCert) won't unify
not_tokenWithValidAuth(Token) :-
  Token = [TokenID,IDCert,_,_,AuthCert] &
    IDCert = [[Id,_,_],[_,_]] &

  (AuthCert = {}
   or
   the(AuthCert,AC) & AC = [_,_,_,BaseCertId,ACTokenID,_,_] &
   (ACTokenID neq TokenID
    or
    BaseCertId neq Id
   )
  ).

not_certIssuerKnown(KeyStore,Certificate) :-
  KeyStore = [IssuerKey,_] &
  Certificate = [Id,_,_] &
    Id = [Issuer] &

%  Issuer nin D & dom(IssuerKey,D)
   comp({[Issuer,Issuer]},IssuerKey,{}).

not_certOK(KeyStore,Certificate) :-
  KeyStore = [IssuerKey,_] &
  Certificate = [Id,_,IsValidatedBy] &
    Id = [Issuer] &

  (not_certIssuerKnown(KeyStore,Certificate)
   or
   M nin IsValidatedBy & [Issuer,M] in IssuerKey
  ).

not_certIssuerIsThisTIS(KeyStore,Certificate) :-
  KeyStore = [_,OwnName] &
  Certificate = [Id,_,_] &
    Id = [Issuer] &

  (OwnName = {}
   or
   Issuer neq M1 & the(OwnName,M1)
  ).

not_authCertOK(KeyStore,Certificate) :-
  not_certIssuerIsThisTIS(KeyStore,Certificate)
  or
  not_certOK(KeyStore,Certificate).

not_userTokenWithOKAuthCert(KeyStore,UserToken,CurrentTime) :-
   UserToken = [CurrentUserToken,_] &

  (CurrentUserToken in {noT, badT}
   or
   CurrentUserToken = goodT(T) & 
     T = [_,IDCert,_,_,AuthCert] &
   (not_tokenWithValidAuth(CurrentUserToken)
    or
    CurrentTime nin ValidityPeriod & 
      the(AuthCert,M3) & M3 = [_,ValidityPeriod,_,_,_,_,_]
    or
    IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] & 
    not_certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy])
    or
    the(AuthCert,M3) & M3 = [Id,ValidityPeriod,IsValidatedBy,_,_,_,_] &
      not_authCertOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy])
   )
  ).

not_currentToken(Token,Now) :-
  Token = [_,IDCert,PrivCert,IandACert,_] &
    IDCert = [[_,ValidityPeriod,_],[_,_]] &
    PrivCert = [_,PCValidityPeriod,_,_,_,_,_] &  
    IandACert = [_,IAValidityPeriod,_,_,_,_] &  

  (not_validToken(Token)
   or
   Now nin M2 & 
     inters(ValidityPeriod,PCValidityPeriod,M1) & 
     inters(M1,IAValidityPeriod,M2)
  ).

not_userTokenOK(KeyStore,UserToken,CurrentTime,Now) :-
  UserToken = [CurrentUserToken,_] &

  (CurrentUserToken in {noT, badT}
   or
   CurrentUserToken = goodT(Token) & 
     Token = [_,IDCert,PrivCert,IandACert,_] &
       IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] &
       PrivCert = [PCId,PCValidityPeriod,PCIsValidatedBy,_,_,_,_] &
       IandACert = [ICId,ICValidityPeriod,ICIsValidatedBy,_,_,_] &
   (not_currentToken(Token,Now)
    or
    Now neq CurrentTime
    or
    not_certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy])
    or
    not_certOK(KeyStore,[PCId,PCValidityPeriod,PCIsValidatedBy])
    or
    not_certOK(KeyStore,[ICId,ICValidityPeriod,ICIsValidatedBy])
   )
  ).

not_validToken(Token) :-
  Token = [TokenID,IDCert,PrivCert,IandACert,_] &
    IDCert = [[Id,_,_],[_,_]] &
    PrivCert = [_,_,_,BaseCertId,PCTokenID,_,_] &
    IandACert = [_,_,_,ICBaseCertId,ICTokenID,_] &

  (BaseCertId neq Id 
   or
   ICBaseCertId neq Id 
   or
   PCTokenID neq TokenID
   or
   ICTokenID neq TokenID
  ).

not_userAllowedEntry(UserToken,Config,CurrentTime) :-
  UserToken = [CurrentUserToken,_] &
  Config = [_,_,_,_,_,EntryPeriod,_,_] &

  (0 > CurrentTime
   or
   CurrentUserToken in {noT, badT}
   or
   goodT(T) = CurrentUserToken & 
     T = [_,_,PrivCert,_,AuthCert] & 
   (not_validToken(T)
    or
    CurrentTime nin M1 & [Role,[Class,M1]] in EntryPeriod &
      PrivCert = [_,_,_,_,_,Role,Clearance] & Clearance = [Class]
   ) &
   (not_tokenWithValidAuth(T)
    or
    CurrentTime nin M2 & [ACRole,[ACClass,M2]] in EntryPeriod &
      the(AuthCert,AC) & 
        AC = [_,_,_,_,_,ACRole,ACClearance] & ACClearance = [ACClass]
    or
    AuthCert = {}
   )
  ).

not_caIdCertPred(IDCert) :- 
  IDCert = [[_,_,IsValidatedBy],[_,SubjectPubK]] &

  IsValidatedBy neq {SubjectPubK}.

not_validEnrol(Enrol) :-
  Enrol = [IdStationCert,IssuerCerts] &

  (IdStationCert nin IssuerCerts
   or
   foreach(X in IssuerCerts, not_caIdCertPred(X))
   or
   delay(
     nforeach(
       [[[Issuer],_,IsValidatedBy],[_,_]] in IssuerCerts,
       [SPK],
       IsValidatedBy neq {} &
       ris([[_,_,IsValidatedBy1],[Subject1,SubjectPubK1]] 
             in IssuerCerts,
           IsValidatedBy1 = {SubjectPubK1} &
           SPK = SubjectPubK1 &
           Issuer = Subject1
          ) neq {},
       the(IsValidatedBy,SPK)
     ),
     false
   )
  ).

not_enrolmentDataOK(Floppy,_KeyStore) :-
  Floppy = [CurrentFloppy,_,_] &

  (CurrentFloppy in {noFloppy, emptyFloppy, badFloppy}
   or
   CurrentFloppy = enrolmentFile(VE) &
   not_validEnrol(VE)
   or
   CurrentFloppy = auditFile(_)
   or
   CurrentFloppy = configFile(_)
  ).

not_adminTokenOK(AdminToken,KeyStore,CurrentTime) :-
  AdminToken = [CurrentAdminToken,_] &

  (CurrentAdminToken in {noT, badT}
   or
   CurrentAdminToken = goodT(T) & T = [_,IDCert,_,_,AuthCert] &
   (not_tokenWithValidAuth(T)
    or
    IDCert = [[Id,ValidityPeriod,IsValidatedBy],[_,_]] &
    not_certOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy]) 
    or
    the(AuthCert,AC) & AC = [Id,ValidityPeriod,IsValidatedBy,_,_,Role,_] &
    (not_authCertOK(KeyStore,[Id,ValidityPeriod,IsValidatedBy])
     or
     Role nin ADMINPRIVILEGE & adminprivilege(ADMINPRIVILEGE)
     or
     CurrentTime nin ValidityPeriod
    )
   )
  ).

% Correct only if Elem is a constant
not_the(Set,Elem) :- Set neq {Elem}.  

% not_the(C,E) iff forall E, C neq {E}
%              iff C = {} or C = {H} & E neq H or C = {A,B/D} & A neq B

not_fingerOK(Finger,_UserToken) :-
  Finger = [CurrentFinger,_] &

  CurrentFinger in {noFP,badFP}.

